@@ -15,7 +15,7 @@ from distributed.utils import get_ip, get_ipv6
 from distributed.utils_test import (gen_test, requires_ipv6, has_ipv6,
                                     get_cert, get_server_ssl_context,
                                     get_client_ssl_context)
-from distributed.utils_test import loop  # flake8: noqa
+from distributed.utils_test import loop  # noqa: F401
 
 from distributed.protocol import (to_serialize, Serialized, serialize,
                                   deserialize)
@@ -712,7 +712,10 @@ def test_inproc_comm_closed_explicit_2():
     comm = yield connect(contact_addr)
     comm.close()
     assert comm.closed()
-    yield gen.sleep(0.01)
+    start = time()
+    while len(listener_errors) < 1:
+        assert time() < start + 1
+        yield gen.sleep(0.01)
     assert len(listener_errors) == 1
 
     with pytest.raises(CommClosedError):
@@ -747,7 +750,7 @@ def check_connect_timeout(addr):
     with pytest.raises(IOError):
         yield connect(addr, timeout=0.15)
     dt = time() - t1
-    assert 0.5 >= dt >= 0.1
+    assert 1 >= dt >= 0.1
 
 
 @gen_test()
@@ -954,8 +957,8 @@ def check_deserialize_roundtrip(addr):
            'ser': Serialized(*serialize(_uncompressible)),
            }
 
-    for deserialize in (True, False):
-        a, b = yield get_comm_pair(addr, deserialize=deserialize)
+    for should_deserialize in (True, False):
+        a, b = yield get_comm_pair(addr, deserialize=should_deserialize)
         yield a.write(msg)
         got = yield b.read()
         yield b.write(got)
@@ -964,7 +967,7 @@ def check_deserialize_roundtrip(addr):
         assert sorted(got) == sorted(msg)
         for k in ('op', 'x'):
             assert got[k] == msg[k]
-        if deserialize:
+        if should_deserialize:
             assert isinstance(got['to_ser'][0], (bytes, bytearray))
             assert isinstance(got['ser'], (bytes, bytearray))
         else:
