@@ -7,6 +7,7 @@ import os
 import psutil
 import shutil
 import threading
+import uuid
 
 from tornado import gen
 from tornado.ioloop import IOLoop, TimeoutError
@@ -42,11 +43,13 @@ class Nanny(ServerNode):
                  ncores=None, loop=None, local_dir=None, services=None,
                  name=None, memory_limit='auto', reconnect=True,
                  validate=False, quiet=False, resources=None, silence_logs=None,
-                 death_timeout=None, preload=(), security=None,
+                 death_timeout=None, preload=(), preload_argv=[], security=None,
                  contact_address=None, listen_address=None, **kwargs):
         if scheduler_file:
             cfg = json_load_robust(scheduler_file)
             self.scheduler_addr = cfg['address']
+        elif scheduler_ip is None and config.get('scheduler-address'):
+            self.scheduler_addr = config['scheduler-address']
         elif scheduler_port is None:
             self.scheduler_addr = coerce_to_address(scheduler_ip)
         else:
@@ -58,6 +61,8 @@ class Nanny(ServerNode):
         self.resources = resources
         self.death_timeout = death_timeout
         self.preload = preload
+        self.preload_argv = preload_argv
+
         self.contact_address = contact_address
         self.memory_terminate_fraction = config.get('worker-memory-terminate', 0.95)
 
@@ -93,7 +98,7 @@ class Nanny(ServerNode):
                                     **kwargs)
 
         if self.memory_limit:
-            pc = PeriodicCallback(self.memory_monitor, 100)
+            pc = PeriodicCallback(self.memory_monitor, 100, io_loop=self.loop)
             self.periodic_callbacks['memory'] = pc
 
         self._listen_address = listen_address
@@ -149,9 +154,11 @@ class Nanny(ServerNode):
 
         logger.info('        Start Nanny at: %r', self.address)
         response = yield self.instantiate()
-        if response == 'OK':
+        if response == 'running':
             assert self.worker_address
             self.status = 'running'
+        else:
+            yield self._close()
 
         self.start_periodic_callbacks()
 
@@ -201,6 +208,7 @@ class Nanny(ServerNode):
                                    silence_logs=self.silence_logs,
                                    death_timeout=self.death_timeout,
                                    preload=self.preload,
+                                   preload_argv=self.preload_argv,
                                    security=self.security,
                                    contact_address=self.contact_address),
                 worker_start_args=(start_arg,),
@@ -211,17 +219,19 @@ class Nanny(ServerNode):
         self.auto_restart = True
         if self.death_timeout:
             try:
-                yield gen.with_timeout(timedelta(seconds=self.death_timeout),
-                                       self.process.start())
+                result = yield gen.with_timeout(
+                        timedelta(seconds=self.death_timeout),
+                        self.process.start()
+                )
             except gen.TimeoutError:
                 yield self._close(timeout=self.death_timeout)
                 raise gen.Return('timed out')
         else:
-            yield self.process.start()
-        raise gen.Return('OK')
+            result = yield self.process.start()
+        raise gen.Return(result)
 
     @gen.coroutine
-    def restart(self, comm=None, timeout=2):
+    def restart(self, comm=None, timeout=2, executor_wait=True):
         start = time()
 
         @gen.coroutine
@@ -242,11 +252,18 @@ class Nanny(ServerNode):
         """ Track worker's memory.  Restart if it goes above 95% """
         if self.status != 'running':
             return
-        memory = psutil.Process(self.process.pid).memory_info().rss
+        process = self.process.process
+        if process is None:
+            return
+        try:
+            proc = psutil.Process(process.pid)
+        except psutil.NoSuchProcess:
+            return
+        memory = proc.memory_info().rss
         frac = memory / self.memory_limit
         if self.memory_terminate_fraction and frac > self.memory_terminate_fraction:
             logger.warning("Worker exceeded 95% memory budget.  Restarting")
-            self.process.process.terminate()
+            process.terminate()
 
     def is_alive(self):
         return self.process is not None and self.process.status == 'running'
@@ -322,13 +339,15 @@ class WorkerProcess(object):
         """
         enable_proctitle_on_children()
         if self.status == 'running':
-            return
+            raise gen.Return(self.status)
         if self.status == 'starting':
             yield self.running.wait()
-            return
+            raise gen.Return(self.status)
 
-        self.init_result_q = mp_context.Queue()
+        self.init_result_q = init_q = mp_context.Queue()
         self.child_stop_q = mp_context.Queue()
+        uid = uuid.uuid4().hex
+
         self.process = AsyncProcess(
             target=self._run,
             kwargs=dict(worker_args=self.worker_args,
@@ -336,7 +355,8 @@ class WorkerProcess(object):
                         worker_start_args=self.worker_start_args,
                         silence_logs=self.silence_logs,
                         init_result_q=self.init_result_q,
-                        child_stop_q=self.child_stop_q),
+                        child_stop_q=self.child_stop_q,
+                        uid=uid),
         )
         self.process.daemon = True
         self.process.set_exit_callback(self._on_exit)
@@ -344,8 +364,18 @@ class WorkerProcess(object):
         self.stopped = Event()
         self.status = 'starting'
         yield self.process.start()
-        if self.status == 'starting':
-            yield self._wait_until_running()
+        msg = yield self._wait_until_connected(uid)
+        if not msg:
+            raise gen.Return(self.status)
+        self.worker_address = msg['address']
+        self.worker_dir = msg['dir']
+        assert self.worker_address
+        self.status = 'running'
+        self.running.set()
+
+        init_q.close()
+
+        raise gen.Return(self.status)
 
     def _on_exit(self, proc):
         if proc is not self.process:
@@ -394,7 +424,7 @@ class WorkerProcess(object):
                 self.on_exit(r)
 
     @gen.coroutine
-    def kill(self, timeout=2):
+    def kill(self, timeout=2, executor_wait=True):
         """
         Ensure the worker process is stopped, waiting at most
         *timeout* seconds before terminating it abruptly.
@@ -411,9 +441,12 @@ class WorkerProcess(object):
         self.status = 'stopping'
 
         process = self.process
-        self.child_stop_q.put({'op': 'stop',
-                               'timeout': max(0, deadline - loop.time()) * 0.8,
-                               })
+        self.child_stop_q.put({
+            'op': 'stop',
+            'timeout': max(0, deadline - loop.time()) * 0.8,
+            'executor_wait': executor_wait,
+        })
+        self.child_stop_q.close()
 
         while process.is_alive() and loop.time() < deadline:
             yield gen.sleep(0.05)
@@ -427,7 +460,7 @@ class WorkerProcess(object):
                 logger.error("Failed to kill worker process: %s", e)
 
     @gen.coroutine
-    def _wait_until_running(self):
+    def _wait_until_connected(self, uid):
         delay = 0.05
         while True:
             if self.status != 'starting':
@@ -438,20 +471,20 @@ class WorkerProcess(object):
                 yield gen.sleep(delay)
                 continue
 
-            if isinstance(msg, Exception):
+            if msg['uid'] != uid:  # ensure that we didn't cross queues
+                continue
+
+            if 'exception' in msg:
+                logger.error("Failed while trying to start worker process: %s",
+                             msg['exception'])
                 yield self.process.join()
                 raise msg
             else:
-                self.worker_address = msg['address']
-                self.worker_dir = msg['dir']
-                assert self.worker_address
-                self.status = 'running'
-                self.running.set()
                 raise gen.Return(msg)
 
     @classmethod
     def _run(cls, worker_args, worker_kwargs, worker_start_args,
-             silence_logs, init_result_q, child_stop_q):  # pragma: no cover
+             silence_logs, init_result_q, child_stop_q, uid):  # pragma: no cover
         from distributed import Worker
 
         try:
@@ -470,9 +503,12 @@ class WorkerProcess(object):
         worker = Worker(*worker_args, **worker_kwargs)
 
         @gen.coroutine
-        def do_stop(timeout):
+        def do_stop(timeout=5, executor_wait=True):
             try:
-                yield worker._close(report=False, nanny=False)
+                yield worker._close(report=False,
+                                    nanny=False,
+                                    executor_wait=executor_wait,
+                                    timeout=timeout)
             finally:
                 loop.stop()
 
@@ -487,8 +523,9 @@ class WorkerProcess(object):
                 except Empty:
                     pass
                 else:
-                    assert msg['op'] == 'stop'
-                    loop.add_callback(do_stop, msg['timeout'])
+                    child_stop_q.close()
+                    assert msg.pop('op') == 'stop'
+                    loop.add_callback(do_stop, **msg)
                     break
 
         t = threading.Thread(target=watch_stop_q, name="Nanny stop queue watch")
@@ -504,11 +541,14 @@ class WorkerProcess(object):
                 yield worker._start(*worker_start_args)
             except Exception as e:
                 logger.exception("Failed to start worker")
-                init_result_q.put(e)
+                init_result_q.put({'uid': uid, 'exception': e})
+                init_result_q.close()
             else:
                 assert worker.address
                 init_result_q.put({'address': worker.address,
-                                   'dir': worker.local_dir})
+                                   'dir': worker.local_dir,
+                                   'uid': uid})
+                init_result_q.close()
                 yield worker.wait_until_closed()
                 logger.info("Worker closed")
 
