@@ -45,8 +45,8 @@ with ignoring(ImportError):
 
     def _fixed_snappy_decompress(data):
         # snappy.decompress() doesn't accept memoryviews
-        if isinstance(data, memoryview):
-            data = data.tobytes()
+        if isinstance(data, (memoryview, bytearray)):
+            data = bytes(data)
         return snappy.decompress(data)
 
     compressions['snappy'] = {'compress': snappy.compress,
@@ -58,9 +58,10 @@ with ignoring(ImportError):
 
     try:
         # try using the new lz4 API
+        import lz4.block
         lz4_compress = lz4.block.compress
         lz4_decompress = lz4.block.decompress
-    except AttributeError as err:
+    except ImportError:
         # fall back to old one
         lz4_compress = lz4.LZ4_compress
         lz4_decompress = lz4.LZ4_uncompress
@@ -72,8 +73,8 @@ with ignoring(ImportError):
         try:
             return lz4_compress(data)
         except TypeError:
-            if isinstance(data, memoryview):
-                return lz4_compress(data.tobytes())
+            if isinstance(data, (memoryview, bytearray)):
+                return lz4_compress(bytes(data))
             else:
                 raise
 
@@ -81,8 +82,8 @@ with ignoring(ImportError):
         try:
             return lz4_decompress(data)
         except (ValueError, TypeError):
-            if isinstance(data, memoryview):
-                return lz4_decompress(data.tobytes())
+            if isinstance(data, (memoryview, bytearray)):
+                return lz4_decompress(bytes(data))
             else:
                 raise
 
