@@ -1,6 +1,7 @@
 from __future__ import print_function, division, absolute_import
 
 from collections import defaultdict
+import datetime
 import logging
 import uuid
 
@@ -32,13 +33,17 @@ class QueueExtension(object):
         self.client_refcount = dict()
         self.future_refcount = defaultdict(lambda: 0)
 
-        self.scheduler.handlers.update({'queue_create': self.create,
-                                        'queue_release': self.release,
-                                        'queue_put': self.put,
-                                        'queue_get': self.get,
-                                        'queue_qsize': self.qsize})
+        self.scheduler.handlers.update({
+            'queue_create': self.create,
+            'queue_put': self.put,
+            'queue_get': self.get,
+            'queue_qsize': self.qsize}
+        )
 
-        self.scheduler.client_handlers['queue-future-release'] = self.future_release
+        self.scheduler.client_handlers.update({
+            'queue-future-release': self.future_release,
+            'queue_release': self.release,
+        })
 
         self.scheduler.extensions['queues'] = self
 
@@ -50,13 +55,18 @@ class QueueExtension(object):
             self.client_refcount[name] += 1
 
     def release(self, stream=None, name=None, client=None):
+        if name not in self.queues:
+            return
+
         self.client_refcount[name] -= 1
         if self.client_refcount[name] == 0:
             del self.client_refcount[name]
-            futures = self.queues[name].queue
+            futures = self.queues[name]._queue
             del self.queues[name]
-            self.scheduler.client_releases_keys(keys=[f.key for f in futures],
-                                                client='queue-%s' % name)
+            self.scheduler.client_releases_keys(
+                    keys=[d['value'] for d in futures if d['type'] == 'Future'],
+                    client='queue-%s' % name
+            )
 
     @gen.coroutine
     def put(self, stream=None, name=None, key=None, data=None, client=None, timeout=None):
@@ -66,6 +76,8 @@ class QueueExtension(object):
             self.scheduler.client_desires_keys(keys=[key], client='queue-%s' % name)
         else:
             record = {'type': 'msgpack', 'value': data}
+        if timeout is not None:
+            timeout = datetime.timedelta(seconds=(timeout))
         yield self.queues[name].put(record, timeout=timeout)
 
     def future_release(self, name=None, key=None, client=None):
@@ -83,15 +95,13 @@ class QueueExtension(object):
             if record['type'] == 'Future':
                 record = record.copy()
                 key = record['value']
-                try:
-                    state = self.scheduler.task_state[key]
-                except KeyError:
-                    state = 'lost'
+                ts = self.scheduler.tasks.get(key)
+                state = ts.state if ts is not None else 'lost'
 
                 record['state'] = state
                 if state == 'erred':
-                    record['exception'] = self.scheduler.exceptions[self.scheduler.exceptions_blame[key]]
-                    record['traceback'] = self.scheduler.tracebacks[self.scheduler.exceptions_blame[key]]
+                    record['exception'] = ts.exception_blame.exception
+                    record['traceback'] = ts.exception_blame.traceback
 
             return record
 
@@ -113,6 +123,8 @@ class QueueExtension(object):
             out = [process(o) for o in out]
             raise gen.Return(out)
         else:
+            if timeout is not None:
+                timeout = datetime.timedelta(seconds=timeout)
             record = yield self.queues[name].get(timeout=timeout)
             record = process(record)
             raise gen.Return(record)
@@ -234,14 +246,11 @@ class Queue(object):
         result = yield self.client.scheduler.queue_qsize(name=self.name)
         raise gen.Return(result)
 
-    def _release(self):
+    def close(self):
         if self.client.status == 'running':  # TODO: can leave zombie futures
             self.client._send_to_scheduler({'op': 'queue_release',
                                             'name': self.name})
 
-    def __del__(self):
-        self._release()
-
     def __getstate__(self):
         return (self.name, self.client.scheduler.address)
 
