@@ -5,11 +5,10 @@ from collections import Iterable, deque
 from contextlib import contextmanager
 from datetime import timedelta
 import functools
-import gc
 import json
 import logging
-import math
 import multiprocessing
+from numbers import Number
 import operator
 import os
 import re
@@ -211,7 +210,8 @@ def sync(loop, func, *args, **kwargs):
     Run coroutine in loop running in separate thread.
     """
     # Tornado's PollIOLoop doesn't raise when using closed, do it ourselves
-    if isinstance(loop, PollIOLoop) and getattr(loop, '_closing', False):
+    if ((isinstance(loop, PollIOLoop) and getattr(loop, '_closing', False)) or
+        (hasattr(loop, 'asyncio_loop') and loop.asyncio_loop._closed)):
         raise RuntimeError("IOLoop is closed")
 
     timeout = kwargs.pop('callback_timeout', None)
@@ -237,7 +237,6 @@ def sync(loop, func, *args, **kwargs):
             thread_state.asynchronous = True
             result[0] = yield make_coro()
         except Exception as exc:
-            logger.exception(exc)
             error[0] = sys.exc_info()
         finally:
             thread_state.asynchronous = False
@@ -249,7 +248,7 @@ def sync(loop, func, *args, **kwargs):
             raise gen.TimeoutError("timed out after %s s." % (timeout,))
     else:
         while not e.is_set():
-            e.wait(1000000)
+            e.wait(10)
     if error[0]:
         six.reraise(*error[0])
     else:
@@ -277,9 +276,10 @@ class LoopRunner(object):
     _lock = threading.Lock()
 
     def __init__(self, loop=None, asynchronous=False):
+        current = IOLoop.current()
         if loop is None:
             if asynchronous:
-                self._loop = IOLoop.current()
+                self._loop = current
             else:
                 # We're expecting the loop to run in another thread,
                 # avoid re-using this thread's assigned loop
@@ -639,13 +639,16 @@ def silence_logging(level, root='distributed'):
     (or keep the existing level if less verbose).
     """
     if isinstance(level, str):
-        level = logging_names[level.upper()]
+        level = getattr(logging, level.upper())
 
-    for name, logger in logging.root.manager.loggerDict.items():
-        if (isinstance(logger, logging.Logger)
-            and logger.name.startswith(root + '.')
-                and logger.level < level):
-            logger.setLevel(level)
+    old = None
+    logger = logging.getLogger(root)
+    for handler in logger.handlers:
+        if isinstance(handler, logging.StreamHandler):
+            old = handler.level
+            handler.setLevel(level)
+
+    return old
 
 
 @memoize
@@ -907,7 +910,7 @@ def ensure_bytes(s):
         return s
     if isinstance(s, memoryview):
         return s.tobytes()
-    if isinstance(s, bytearray) or PY2 and isinstance(s, buffer):  # flake8: noqa
+    if isinstance(s, bytearray) or PY2 and isinstance(s, buffer):  # noqa: F821
         return bytes(s)
     if hasattr(s, 'encode'):
         return s.encode()
@@ -984,19 +987,19 @@ def open_port(host=''):
 
 
 def import_file(path):
-    """ Loads modules for a file (.py, .pyc, .zip, .egg) """
+    """ Loads modules for a file (.py, .zip, .egg) """
     directory, filename = os.path.split(path)
     name, ext = os.path.splitext(filename)
     names_to_import = []
     tmp_python_path = None
 
-    if ext in ('.py', '.pyc'):
+    if ext in ('.py'):  # , '.pyc'):
         if directory not in sys.path:
             tmp_python_path = directory
         names_to_import.append(name)
-        # Ensures that no pyc file will be reused
+    if ext == '.py':  # Ensure that no pyc file will be reused
         cache_file = cache_from_source(path)
-        if os.path.exists(cache_file):
+        with ignoring(OSError):
             os.remove(cache_file)
     if ext in ('.egg', '.zip'):
         if path not in sys.path:
@@ -1130,6 +1133,70 @@ def parse_bytes(s):
     return int(result)
 
 
+timedelta_sizes = {
+        's': 1,
+        'ms': 1e-3,
+        'us': 1e-6,
+        'ns': 1e-9,
+        'm': 60,
+        'h': 3600,
+        'd': 3600 * 24,
+}
+
+tds2 = {
+        'second': 1,
+        'minute': 60,
+        'hour': 60 * 60,
+        'day': 60 * 60 * 24,
+        'millisecond': 1e-3,
+        'microsecond': 1e-6,
+        'nanosecond': 1e-9,
+}
+tds2.update({k + 's': v for k, v in tds2.items()})
+timedelta_sizes.update(tds2)
+timedelta_sizes.update({k.upper(): v for k, v in timedelta_sizes.items()})
+
+
+def parse_timedelta(s, default='seconds'):
+    """ Parse timedelta string to number of seconds
+
+    Examples
+    --------
+    >>> parse_timedelta('3s')
+    3
+    >>> parse_timedelta('3.5 seconds')
+    3.5
+    >>> parse_timedelta('300ms')
+    0.3
+    >>> parse_timedelta(timedelta(seconds=3))  # also supports timedeltas
+    3
+    """
+    if isinstance(s, timedelta):
+        return s.total_seconds()
+    if isinstance(s, Number):
+        s = str(s)
+    s = s.replace(' ', '')
+    if not s[0].isdigit():
+        s = '1' + s
+
+    for i in range(len(s) - 1, -1, -1):
+        if not s[i].isalpha():
+            break
+    index = i + 1
+
+    prefix = s[:index]
+    suffix = s[index:] or default
+
+    n = float(prefix)
+
+    multiplier = timedelta_sizes[suffix.lower()]
+
+    result = n * multiplier
+    if int(result) == result:
+        result = int(result)
+    return result
+
+
 def asciitable(columns, rows):
     """Formats an ascii table for given columns and rows.
 
@@ -1153,7 +1220,7 @@ def asciitable(columns, rows):
 
 
 if PY2:
-    def nbytes(frame, _bytes_like=(bytes, bytearray, buffer)):
+    def nbytes(frame, _bytes_like=(bytes, bytearray, buffer)):  # noqa: F821
         """ Number of bytes of a frame or memoryview """
         if isinstance(frame, _bytes_like):
             return len(frame)
@@ -1256,53 +1323,6 @@ class DequeHandler(logging.Handler):
             inst.clear()
 
 
-class ThrottledGC(object):
-    """Wrap gc.collect to protect against excessively repeated calls.
-
-    Allows to run throttled garbage collection in the workers as a
-    countermeasure to e.g.: https://github.com/dask/zict/issues/19
-
-    collect() does nothing when repeated calls are so costly and so frequent
-    that the thread would spend more than max_in_gc_frac doing GC.
-
-    warn_if_longer is a duration in seconds (10s by default) that can be used
-    to log a warning level message whenever an actual call to gc.collect()
-    lasts too long.
-    """
-    def __init__(self, max_in_gc_frac=0.05, warn_if_longer=1, logger=None):
-        self.max_in_gc_frac = max_in_gc_frac
-        self.warn_if_longer = warn_if_longer
-        self.last_collect = time()
-        self.last_gc_duration = 0
-        self.logger = logger if logger is not None else _logger
-
-    def collect(self):
-        # In case of non-monotonicity in the clock, assume that any Python
-        # operation lasts at least 1e-6 second.
-        collect_start = time()
-        elapsed = max(collect_start - self.last_collect, 1e-6)
-        if self.last_gc_duration / elapsed < self.max_in_gc_frac:
-            self.logger.debug("Calling gc.collect(). %0.3fs elapsed since "
-                              "previous call.", elapsed)
-            gc.collect()
-            self.last_collect = collect_start
-            self.last_gc_duration = max(time() - collect_start, 1e-6)
-            if self.last_gc_duration > self.warn_if_longer:
-                self.logger.warning("gc.collect() took %0.3fs. This is usually"
-                                    " a sign that the some tasks handle too"
-                                    " many Python objects at the same time."
-                                    " Rechunking the work into smaller tasks"
-                                    " might help.",
-                                    self.last_gc_duration)
-            else:
-                self.logger.debug("gc.collect() took %0.3fs",
-                                  self.last_gc_duration)
-        else:
-            self.logger.debug("gc.collect() lasts %0.3fs but only %0.3fs "
-                              "elapsed since last call: throttling.",
-                              self.last_gc_duration, elapsed)
-
-
 def fix_asyncio_event_loop_policy(asyncio):
     """
     Work around https://github.com/tornadoweb/tornado/issues/2183
@@ -1325,6 +1345,16 @@ def fix_asyncio_event_loop_policy(asyncio):
     asyncio.set_event_loop_policy(PatchedDefaultEventLoopPolicy())
 
 
+def reset_logger_locks():
+    """ Python 2's logger's locks don't survive a fork event
+
+    https://github.com/dask/distributed/issues/1491
+    """
+    for name in logging.Logger.manager.loggerDict.keys():
+        for handler in logging.getLogger(name).handlers:
+            handler.createLock()
+
+
 # Only bother if asyncio has been loaded by Tornado
 if 'asyncio' in sys.modules:
     fix_asyncio_event_loop_policy(sys.modules['asyncio'])
