@@ -1,5 +1,6 @@
 from __future__ import print_function, division, absolute_import
 
+import collections
 from contextlib import contextmanager
 from datetime import timedelta
 import functools
@@ -21,7 +22,6 @@ import textwrap
 import threading
 from time import sleep
 import uuid
-import warnings
 import weakref
 
 try:
@@ -29,7 +29,6 @@ try:
 except ImportError:
     ssl = None
 
-import psutil
 import pytest
 import six
 
@@ -39,15 +38,14 @@ from tornado import gen, queues
 from tornado.gen import TimeoutError
 from tornado.ioloop import IOLoop
 
-from .compatibility import WINDOWS, PY3
+from .compatibility import PY3, iscoroutinefunction, Empty
 from .config import config, initialize_logging
 from .core import connect, rpc, CommClosedError
 from .metrics import time
-from .nanny import Nanny
 from .proctitle import enable_proctitle_on_children
 from .security import Security
-from .utils import (ignoring, log_errors, sync, mp_context, get_ip, get_ipv6,
-                    DequeHandler)
+from .utils import (ignoring, log_errors, mp_context, get_ip, get_ipv6,
+                    DequeHandler, reset_logger_locks)
 from .worker import Worker, TOTAL_MEMORY
 
 
@@ -89,6 +87,7 @@ def loop():
         orig_start = loop.start
         is_stopped = threading.Event()
         is_stopped.set()
+
         def start():
             is_stopped.clear()
             try:
@@ -143,7 +142,7 @@ def pristine_loop():
     finally:
         try:
             loop.close(all_fds=True)
-        except ValueError:
+        except (KeyError, ValueError):
             pass
         IOLoop.clear_instance()
         IOLoop.clear_current()
@@ -152,14 +151,21 @@ def pristine_loop():
 @contextmanager
 def mock_ipython():
     import mock
+    from distributed._ipython_utils import remote_magic
     ip = mock.Mock()
     ip.user_ns = {}
     ip.kernel = None
 
-    def get_ip(): return ip
+    def get_ip():
+        return ip
+
     with mock.patch('IPython.get_ipython', get_ip), \
             mock.patch('distributed._ipython_utils.get_ipython', get_ip):
         yield ip
+    # cleanup remote_magic client cache
+    for kc in remote_magic._clients.values():
+        kc.stop_channels()
+    remote_magic._clients.clear()
 
 
 def nodebug(func):
@@ -277,10 +283,20 @@ def slowidentity(*args, **kwargs):
         return args
 
 
+def run_for(duration, timer=time):
+    """
+    Burn CPU for *duration* seconds.
+    """
+    deadline = timer() + duration
+    while timer() <= deadline:
+        pass
+
+
 # This dict grows at every varying() invocation
-_varying_dict = {}
+_varying_dict = collections.defaultdict(int)
 _varying_key_gen = itertools.count()
 
+
 class _ModuleSlot(object):
     def __init__(self, modname, slotname):
         self.modname = modname
@@ -301,7 +317,6 @@ def varying(items):
     # and not a copy.
     slot = _ModuleSlot(__name__, '_varying_dict')
     key = next(_varying_key_gen)
-    _varying_dict[key] = 0
 
     def func():
         dct = slot.get()
@@ -350,7 +365,7 @@ if sys.version_info >= (3, 5):
             await gen.sleep(delay)
             return x + 1
         """)
-    assert asyncinc  # flake8: noqa
+    assert asyncinc  # noqa: F821
 else:
     asyncinc = None
 
@@ -410,6 +425,7 @@ def run_scheduler(q, nputs, **kwargs):
 def run_worker(q, scheduler_q, **kwargs):
     from distributed import Worker
 
+    reset_logger_locks()
     with log_errors():
         with pristine_loop() as loop:
             scheduler_addr = scheduler_q.get()
@@ -456,6 +472,7 @@ def check_active_rpc(loop, active_rpc_timeout=1):
     # This would happen especially if a non-localhost address is used,
     # as Nanny does.
     # (*) (example: gather_from_workers())
+
     def fail():
         pytest.fail("some RPCs left active by test: %s"
                     % (sorted(set(rpc.active) - active_before)))
@@ -512,8 +529,11 @@ def cluster(nworkers=2, nanny=False, worker_kwargs={}, active_rpc_timeout=1,
 
             for worker in workers:
                 worker['proc'].start()
-            for worker in workers:
-                worker['address'] = worker['queue'].get()
+            try:
+                for worker in workers:
+                    worker['address'] = worker['queue'].get(timeout=5)
+            except Empty:
+                raise pytest.xfail.Exception("Worker failed to start in test")
 
             saddr = scheduler_q.get()
 
@@ -604,7 +624,10 @@ def gen_test(timeout=10):
     def _(func):
         def test_func():
             with pristine_loop() as loop:
-                cor = gen.coroutine(func)
+                if iscoroutinefunction(func):
+                    cor = func
+                else:
+                    cor = gen.coroutine(func)
                 try:
                     loop.run_sync(cor, timeout=timeout)
                 finally:
@@ -635,9 +658,12 @@ def start_cluster(ncores, scheduler_addr, loop, security=None,
     yield [w._start(ncore[0]) for ncore, w in zip(ncores, workers)]
 
     start = time()
-    while len(s.ncores) < len(ncores):
+    while (len(s.workers) < len(ncores) or
+           any(comm.comm is None for comm in s.worker_comms.values())):
         yield gen.sleep(0.01)
         if time() - start > 5:
+            yield [w._close(timeout=1) for w in workers]
+            yield s.close(fast=True)
             raise Exception("Cluster creation timeout")
     raise gen.Return((s, workers))
 
@@ -677,7 +703,10 @@ def gen_cluster(ncores=[('127.0.0.1', 1), ('127.0.0.1', 2)],
         start
         end
     """
-    worker_kwargs = merge({'memory_limit': TOTAL_MEMORY}, worker_kwargs)
+    config['nanny-start-timeout'] = '5s'
+    config['connect-timeout'] = '5s'
+    worker_kwargs = merge({'memory_limit': TOTAL_MEMORY, 'death_timeout': 5},
+                          worker_kwargs)
 
     def _(func):
         if not iscoroutinefunction(func):
@@ -697,10 +726,16 @@ def gen_cluster(ncores=[('127.0.0.1', 1), ('127.0.0.1', 2)],
                 with check_active_rpc(loop, active_rpc_timeout):
                     @gen.coroutine
                     def coro():
-                        s, ws = yield start_cluster(
-                            ncores, scheduler, loop, security=security,
-                            Worker=Worker, scheduler_kwargs=scheduler_kwargs,
-                            worker_kwargs=worker_kwargs)
+                        for i in range(5):
+                            try:
+                                s, ws = yield start_cluster(
+                                    ncores, scheduler, loop, security=security,
+                                    Worker=Worker, scheduler_kwargs=scheduler_kwargs,
+                                    worker_kwargs=worker_kwargs)
+                            except Exception as e:
+                                logger.error("Failed to start gen_cluster, retryng", exc_info=True)
+                            else:
+                                break
                         workers[:] = ws
                         args = [s] + workers
                         if client:
@@ -708,19 +743,23 @@ def gen_cluster(ncores=[('127.0.0.1', 1), ('127.0.0.1', 2)],
                                              asynchronous=True)
                             args = [c] + args
                         try:
-                            result = yield func(*args)
-                            # for w in workers:
-                            #     assert not w._comms
+                            future = func(*args)
+                            if timeout:
+                                future = gen.with_timeout(timedelta(seconds=timeout),
+                                                          future)
+                            result = yield future
+                            if s.validate:
+                                s.validate_state()
                         finally:
                             if client:
-                                yield c._close()
+                                yield c._close(fast=s.status == 'closed')
                             yield end_cluster(s, workers)
                             _globals.clear()
                             _globals.update(old_globals)
 
                         raise gen.Return(result)
 
-                    result = loop.run_sync(coro, timeout=timeout)
+                    result = loop.run_sync(coro, timeout=timeout * 2 if timeout else timeout)
 
             for w in workers:
                 if getattr(w, 'data', None):
@@ -1035,6 +1074,17 @@ def new_config(new_config):
         initialize_logging(config)
 
 
+@contextmanager
+def new_environment(changes):
+    saved_environ = os.environ.copy()
+    os.environ.update(changes)
+    try:
+        yield
+    finally:
+        os.environ.clear()
+        os.environ.update(saved_environ)
+
+
 @contextmanager
 def new_config_file(c):
     """
